from typing import Callable
from typing import Iterator
from typing import Any

class Callbacks[T: Callable]:
    """
    Callbacks is a class that manages a group of callback objects. It is responsible for
    calling the callback objects and returning their results.

    Args:
        *callbacks: A sequence of callback objects.

    Example:

        .. code-block:: python
        class Flushable:
            ## Supose for example that you need to implement a flush method in all callbacks
            ## you pass to a function that will call a callback

            def flush(self):
                print(f'flushing')
                        
        class SomeCallback(Flushable):
            def __call__(self, *args, **kwargs) -> Any:
                return args
                
        class OtherCallback(Flushable):
            def __call__(self, *args, **kwargs) -> Any:
                return kwargs

        # Create an instance of Callbacks with generic type Flushable
        callbacks = Callbacks[Flushable](SomeCallback(), OtherCallback())

        for result in callbacks(1, 2, 3, a=4, b=5, c=6):
            print(result) # (1, 2, 3) 
                          # {'a': 4, 'b': 5, 'c': 6} 
            
        for callback in callbacks:
            callback.flush() # type hinting will work here
    """

    def __init__(self, *callbacks: T):
        """
        Args:
            *callbacks: A sequence of callback objects.
        """
        self.callbacks = callbacks

    def __call__(self, *args, **kwargs) -> tuple[Any]:
        """
        Calls each callback object with the given arguments and keyword arguments.

        Args:
            *args: A sequence of arguments.
            **kwargs: A dictionary of keyword arguments.

        Returns:
            A tuple of results from each callback object.
        """
        return tuple(callback(*args, **kwargs) for callback in self.callbacks)
    
    def __iter__(self) -> Iterator[T]:
        """
        Returns an iterator over the callback objects.
        """
        return iter(self.callbacks)


from typing import Callable
from typing import Any
from inspect import signature

from fast_depends import inject
from fast_depends import Provider

class Command:
    """
    Represents a base class for commands. Commands are objects that encapsulate
    the data required to perform a specific action or operation.

    This class is intended to be subclassed to define concrete command types
    with the required attributes.
    """
    ...
    def execute(self):
        """
        Executes the command. This method may be overridden by subclasses to provide
        a default implementation for handling the command.
        """
        raise NotImplementedError(f"Not handler nor execute method found for command {self.__class__.__name__}")


class Commands:
    """
    Manages the registration and handling of commands and their associated handlers.

    Attributes:
        key_generator (Callable[[str], str]): A function used to generate keys for commands 
            based on their class names. Defaults to an identity function.
        types (dict[str, type[Command]]): A dictionary mapping command keys to their types.
        handlers (dict[str, Callable[[Command], None]]): A dictionary mapping command keys 
            to their respective handler functions.

    Methods:
        handle(command: Command) -> None:
            Handles a command by invoking its registered handler. Raises a KeyError if 
            no handler is found for the command's key.

        register(command_type: type[Command], handler: Callable[[Command], None]) -> None:
            Registers a command type and its corresponding handler function.

        handler(wrapped: Callable[[Command], Any]) -> Callable:
            Decorator for registering a handler function for one or more command types.

    Example:
        .. code-block:: python
        from dataclasses import dataclass
        from pybondi import Commands

        commands = Commands()
        db = {}  # Database

        @dataclass
        class CreateUser(Command):
            id: str
            name: str

        @dataclass
        class UpdateUser(Command):
            id: str
            name: str

        @commands.handler
        def put_user(command: CreateUser | UpdateUser):
            db[command.id] = command.name

        commands.handle(CreateUser(id='1',name='Alice'))
        commands.handle(UpdateUser(id='1',name='Bob'))
        print(db)  # Output: {'1': 'Bob'}
    """
        
    def __init__(self, key_generator: Callable[[str], str] = lambda name: name, provider: Provider = None, cast_dependency: bool = True):
        """
        Initializes the Commands instance.

        Args:
            key_generator (Callable[[str], str], optional): A function to generate command keys 
                from class names. Defaults to a lambda returning the class name unchanged.
            provider (Provider, optional): A provider instance to use for dependency injection.
            cast_dependency (bool, optional): Whether to cast the dependencies to the expected type.
        """
        self.cast_dependency = cast_dependency
        self.provider = provider or Provider()
        self.key_generator = key_generator
        self.types = dict[str, type[Command]]()
        self.handlers = dict[str, Callable[[Command], None]]()

    def handle(self, command: Command) -> None:
        """
        Handles a command by invoking its associated handler.

        Args:
            command (Command): The command to handle.

        Raises:
            KeyError: If no handler is found for the command's key.
        """
        key = self.key_generator(command.__class__.__name__)
        handler = self.handlers.get(key, None)
        handler(command) if handler else command.execute()
    
    def register(self, command_type: type[Command], handler: Callable[[Command], None]):
        """
        Registers a command type and its handler.

        Args:
            command_type (type[Command]): The type of the command to register.
            handler (Callable[[Command], None]): The handler function for the command.
        """
        key = self.key_generator(command_type.__name__)
        self.types[key] = command_type
        self.handlers[key] = inject(handler, dependency_overrides_provider=self.provider, cast=self.cast_dependency)

    def handler(self, wrapped: Callable[[Command], Any]):
        """
        Decorator to register a handler function for one or more command types.

        Args:
            wrapped (Callable[[Command], Any]): The handler function to register.

        Returns:
            Callable: The original handler function, unmodified.

        Notes:
            If the handler function is annotated with a union of command types,
            all of those types will be registered with the same handler.
        """
        function_signature = signature(wrapped)
        parameter = next(iter(function_signature.parameters.values()))
        if hasattr(parameter.annotation, '__args__'):
            for message_type in getattr(parameter.annotation, '__args__'):
                self.register(message_type, wrapped)
        else:
            message_type = parameter.annotation
            self.register(message_type, wrapped)
        return wrapped


from typing import Callable
from typing import Any
from fast_depends import Provider
from pybondi.publisher import Publisher, Message
from pybondi.commands import Commands, Command
from pybondi.events import Events, Event

class Service:
    """
    The Service class groups a Publisher, Commands and Events instances to provide a single interface for
    creating a service.

    Example:
        .. code-block:: python
    
        from pybondi import Service
        from pybondi import Depends

        service = Service()
        
        def some_dependency():
            raise NotImplementedError

        @service.handler
        def handle_some_cmd(command, dps = Depends(some_dependency)):
            ...

        @service.consumer
        def consume_some_event(event, dps = Depends(some_dependency)):
            ...

        @service.subscriber('topic-1', 'topic-2')
        def on_published(message, dps = Depends(some_dependency)):
            ...

        def concrete_dependency():
            return {}
            
        service.dependency_overrides[some_dependency] = concrete_dependency 
    """
    def __init__(self, key_generator: Callable[[str], str] = lambda name: name, provider: Provider = None, cast_dependency: bool = True):
        """
        Initializes a new instance of the Service class.

        Args:
            key_generator (Callable[[str], str], optional): A function that generates a key from a name. Defaults to lambda name: name.
            provider (Provider, optional): A provider instance to use for dependency injection. Defaults to None.
            cast_dependency (bool, optional): Whether to cast the dependencies to the expected type. Defaults to True.
        
        Notes:
            The key_generator function is used to generate a key from a name. This key is used to store the function in the commands and
            events maps. The default key_generator function simply returns the name unchanged. This is useful when you want to use some
            naming conventions like kebab-case or snake_case for the command and event names.
        """

        self.provider = provider or Provider()
        self.publisher = Publisher(self.provider, cast_dependency)
        self.commands = Commands(key_generator, self.provider, cast_dependency)
        self.events = Events(key_generator, self.provider, cast_dependency)

    @property
    def dependency_overrides(self):
        return self.provider.dependency_overrides
    
    def handle(self, command: Command) -> None:
        """
        Handles a command by invoking its associated handler.

        Args:
            command (Command): The command to handle.

        Raises:
            KeyError: If no handler is found for the command's key.
        """
        return self.commands.handle(command)

    def consume(self, event: Event) -> None:
        """
        Consumes an event by invoking its registered consumer functions.

        Args:
            event (Event): The event to consume.
        """
        return self.events.consume(event)
    
    def publish(self, topic: str, message: Message) -> None:
        """
        Publishes a message to a topic.

        Args:
            topic (str): The topic to publish the message to.
            message (Message): The message to publish.
        """
        return self.publisher.publish(topic, message)   

    def handler(self, wrapped: Callable[[Command], Any]):
        """
        Decorator to register a handler function for one or more command types.

        Args:
            wrapped (Callable[[Command], Any]): The handler function to register.

        Returns:
            Callable: The original handler function, unmodified.

        Notes:
            If the handler function is annotated with a union of command types,
            all of those types will be registered with the same handler.
        """
        return self.commands.handler(wrapped)
    
    def consumer(self, wrapped: Callable[[Event], Any]):
        """
        Decorator for registering a consumer function for one or more event types.

        Args:
            wrapped (Callable[[Event], Any]): The consumer function to register.

        Returns:
            Callable: The original handler function, unmodified.

        Note:
            If the handler function is annotated with a union of event types,
            all of those types will be registered with the same consumer.    
        """
        return self.events.consumer(wrapped)
    
    def subscriber(self, *topics: str):
        '''
        A decorator that can be used to subscribe a callback to a topic. 
        The callback will be called when a message is published to the topic.

        Args:
            *topics: The topics to subscribe to.

        Returns:
            A decorator that can be used to subscribe a callback to a topic.
        '''
        return self.publisher.subscriber(*topics)


from pybondi.commands import Command as Command
from pybondi.events import Event as Event
from pybondi.publisher import Message as Message
from pybondi.publisher import Publisher as Publisher
from pybondi.session import Session as Session
from pybondi.service import Service as Service
from fast_depends import Depends as Depends


from typing import Callable
from typing import Any
from inspect import signature

from fast_depends import Provider
from fast_depends import inject

class Event:
    """
    Represents a base class for events. Events are objects that encapsulate
    the data associated with a specific occurrence or notification.

    This class is intended to be subclassed to define concrete event types
    with the required attributes.
    """
    ...

class Events:
    """
    Manages the registration and handling of events and their associated consumers.

    Attributes:
        key_generator (Callable[[str], str]): A function used to generate keys for events 
            based on their class names. Defaults to an identity function.
        types (dict[str, type[Event]]): A dictionary mapping event keys to their types.
        consumers (dict[str, set[Callable[[Event], None]]]): A dictionary mapping event keys 
            to a set of consumer functions.

    Methods:
        consume(event: Event) -> None:
            Consumes an event by invoking its registered consumer functions.

        register(event_type: type[Event], consumer: Callable[[Event], None]) -> None:
            Registers an event type and its corresponding consumer function.

        consumer(wrapped: Callable[[Event], Any]) -> Callable:
            Decorator for registering a consumer function for one or more event types.


    Example:
        .. code-block:: python

        from dataclasses import dataclass
        from pybondi import Events

        events = Events()
        notifications = []  # List of notifications

        @dataclass
        class UserCreated(Event):
            id: str
            name: str

        @dataclass
        class UserUpdated(Event):
            id: str
            name: str

        @events.consumer
        def on_user_put(event: UserCreated | UserUpdated):
            notifications.append(event)

        @events.consumer
        def on_user_updated(event: UserUpdated):
            print(f"User {event.id} updated with name {event.name}")
        
        events.consume(UserCreated(id="1", name="Alice"))
        events.consume(UserUpdated(id="1", name="Bob")) # Output: User 1 updated with name Bob
        print(notifications)  # Output: [UserCreated(id='1', name='Alice'), UserUpdated(id='1', name='Bob')]
    """

    def __init__(self, key_generator: Callable[[str], str] = lambda name: name, provider: Provider = None, cast_dependency: bool = True):
        """
        Initializes the Events instance.

        Args:
            key_generator (Callable[[str], str], optional): A function to generate event keys 
                from class names. Defaults to a lambda returning the class name unchanged.
            provider (Provider, optional): A provider instance to use for dependency injection.
            cast_dependency (bool, optional): Whether to cast dependencies to the expected type.
        """
        self.cast_dependency = cast_dependency
        self.provider = provider or Provider()
        self.key_generator = key_generator
        self.types = dict[str, type[Event]]()
        self.consumers = dict[str, set[Callable[[Event], None]]]()

    def consume(self, event: Event) -> None:
        """
        Consumes an event by invoking its registered consumer functions.

        Args:
            event (Event): The event to consume.
        """
        key = self.key_generator(event.__class__.__name__)
        for consumer in self.consumers.get(key, set()):
            consumer(event)
    
    def register(self, event_type: type[Event], consumer: Callable[[Event], None]):
        """
        Registers an event type and its corresponding consumer function.

        Args:
            event_type (type[Event]): The type of the event to register.
            consumer (Callable[[Event], None]): The consumer function for the event.
        """


        key = self.key_generator(event_type.__name__)
        self.types[key] = event_type
        self.consumers.setdefault(key, set()).add(inject(consumer, dependency_overrides_provider=self.provider, cast=self.cast_dependency))

    def consumer(self, wrapped: Callable[[Event], Any]):
        """
        Decorator for registering a consumer function for one or more event types.

        Args:
            wrapped (Callable[[Event], Any]): The consumer function to register.

        Returns:
            Callable: The original handler function, unmodified.

        Note:
            If the handler function is annotated with a union of event types,
            all of those types will be registered with the same consumer.    
        """
        function_signature = signature(wrapped)
        parameter = next(iter(function_signature.parameters.values()))
        if hasattr(parameter.annotation, '__args__'):
            for message_type in getattr(parameter.annotation, '__args__'):
                self.register(message_type, wrapped)
        else:
            message_type = parameter.annotation
            self.register(message_type, wrapped)
        return wrapped


from typing import Protocol
from typing import Callable
from typing import Optional
from logging import getLogger
from inspect import signature

logger = getLogger(__name__)

class Resource(Protocol):
    """
    Represents an interface (Protocol) for resources. Resources are objects that can handle
    transactions and provide access to data or services.
    """
    ...
    def begin(self):
        """
        Begins a transaction. This method should be called before any other
        transactional methods are invoked.
        """
        ...

    def commit(self):
        """
        Commits the current transaction. This method should be called after all
        transactional methods have been invoked successfully.
        """
        ...

    def rollback(self):
        """
        Rolls back the current transaction. This method should be called if an
        error occurs during the transaction.
        """
        ...

    def close(self):
        """
        Closes the resource. This method should be called after all transactions
        have been completed.
        """
        ...

class Session:
    """
    Represents a session for managing transactions with a resource. Sessions are used
    to group multiple operations into a single transaction that can be committed or
    rolled back as a unit.

    Custom exception handlers can be registered to handle specific exceptions that
    occur during the session to avoid rolling back the transaction. This is useful
    for example for early stopping in loops.

    Attributes:
        *args: Variable length argument list with resource instances. 

    Methods:
        begin(self) -> None:
            Begins a transaction with the resource.

        commit(self) -> None:
            Commits the current transaction with the resource.

        rollback(self) -> None:
            Rolls back the current transaction with the resource.

        close(self) -> None:
            Closes the session and releases the resource.

    Example:

        .. code-block:: python

        from pybondi import Session
        
        with Session(repository, publisher, logger) as session:
            repository.add(model)
            publisher.publish(event)
            session.commit() # Session will commit all resources

            repository.add(another_model)
            raise Exception("Something went wrong") # Session will rollback all resources
                                                    # to the state where it was committed
    """
    def __init__(self, *args: Resource):
        """
        Initializes a new session with the given resources.

        Args:
            *args: Variable length argument list with resource instances.
        """
        self.resources = args
        self.handler: dict[type[BaseException], Callable[[Optional[BaseException]], bool]] = {}

    def on(self, exception: type[BaseException]):
        """
        Registers an exception handler for the given exception type.

        Args:
            exception: The type of exception to handle.

        Returns:
            A decorator that registers the exception handler.
        """
        def decorator(handler: Callable[[Optional[BaseException]], bool]):
            self.handler[exception] = handler
        return decorator

    def __enter__(self):
        """
        Begins a transaction with the resource. This method should be called
        """
        for resource in self.resources:
            resource.begin()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Handles the session's lifecycle, including invoking exception handlers if an exception occurs.
        if a handler exists for the exception type, non exception will be raised else the exception will be raised.
        if the exception handler returns True, the transaction will be committed, otherwise it will be rolled back.

        Args:
            exc_type: The type of the exception raised.
            exc_value: The exception instance raised.
            traceback: The traceback object.

        """
        if exc_type:
            result = False
            commit = False
            if exc_type in self.handler:
                result = True
                try:
                    handler_signature = signature(self.handler[exc_type])
                    if len(handler_signature.parameters) == 0:
                        commit = self.handler[exc_type]()
                    else:
                        commit = self.handler[exc_type](exc_value)
                except Exception as error:
                    logger.error(f"Error while handling exception: {error}")
                    commit = False
                    result = False
        else:
            result = True
            commit = True

        if commit is True:
            for resource in self.resources:
                resource.commit()
        else:
            for resource in self.resources:
                resource.rollback()

        for resource in self.resources:
            resource.close()
        return result
    
    def commit(self):
        """
        Commits the current transaction with the resource.
        """
        for resource in self.resources:
            resource.commit()

    def rollback(self):
        """
        Rolls back the current transaction with the resource.
        """
        for resource in self.resources:
            resource.rollback()


from typing import Callable
from typing import Any
from fast_depends import Provider
from fast_depends import inject
from fast_depends import Depends as Depends

class Message:
    """
    Represents a base class for messages. Messages are objects that should be
    used to communicate data to components to the outside world.

    This class is intended to be subclassed to define concrete message types
    with the required attributes.
    """

class Publisher:
    """
    Represents a base class for publishers. Publishers are objects that should
    be used to publish messages to subscribers.

    Pydantic validation is performed on the type of the message if a type hint
    is provided. Dependency injection is supported for subscribers.

    Attributes:
        subscribers: A dictionary that maps topics to a set of subscribers
            that are interested in messages published to that topic.

    Methods:
        subscribe: A decorator that can be used to subscribe a callback to a
            topic. The callback will be called when a message is published to
            the topic.
        publish: Publishes a message to all subscribers

    Example:
        .. code-block:: python

        from pybondi import Depends
        from pybondi import Publisher

        publisher = Publisher()
        db = []

        def get_db(): # Define dependencies like in FastAPI
            return db
        
        @publisher.subscriber('topic-1', 'topic-2')
        def callback(message):
            print(message)

        @publisher.subscriber('topic-2')
        def second_callback(message, db = Depends(get_db)): # Dependency injection can be used
            print(message)
            db.append(message)
            
        publisher.publish('topic-2', 'Hello world!') # Will print 'Hello world!' twice
        print(db) # Will print ['Hello world!']
    """
    def __init__(self, provider: Provider = None, cast_dependency: bool = True):
        """
        Args:
            provider (Provider, optional): A provider instance to use for dependency injection. Defaults to None.
            cast_dependency (bool, optional): Whether to cast the dependencies to the expected type. Defaults to True.
        """
        self.cast_dependency = cast_dependency
        self.provider = provider or Provider()
        self.subscribers = dict[str, set[Callable[[Message | Any], None]]]()

    def subscriber(self, *topics: str):
        '''
        A decorator that can be used to subscribe a callback to a topic. 
        The callback will be called when a message is published to the topic.

        Args:
            *topics: The topics to subscribe to.

        Returns:
            A decorator that can be used to subscribe a callback to a topic.
        '''
        def decorator(callback: Callable[[Message | Any], None]):
            for topic in topics:
                if topic not in self.subscribers:
                    self.subscribers[topic] = set()
                injected = inject(callback, dependency_overrides_provider=self.provider, cast=self.cast_dependency)
                self.subscribers[topic].add(injected)
            return callback
        return decorator
    
    def publish(self, topic: str, message: Message | Any):
        """
        Publishes a message to all subscribers. Can be a Message
        or any other object.

        Args:
            topic (str): The topic to publish the message to.
            message (Message | Any): The message to publish.
        """
        for callback in self.subscribers.get(topic, []):
            callback(message)